/***************************************************************************
 *   Copyright (C) 2007 by Francesc Burrull i Mestres.                     *
 *   francesc.burrull@upct.es                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _W2LAN_H_
#include "w2lan.h"
#endif


long mac_addr_sys ( u_int8_t *addr, unsigned char * dvce );



int main(int argc, char *argv[]) {

 /* Beginning of parameter processing */

 /* don't let getopt() write to stderr */
 opterr = 0;

 /* To store getopt results from function main */
 char c;

 /* Device to be opened */
 char * device;

 /* libpcap error handling buffer */
 char errbuf[PCAP_ERRBUF_SIZE];

 /* data link type */
 int datalink = 0;

 /* promiscuous mode = 1 */
 int promisc_flag = 1;

 /* local packet capture descriptor */
 pcap_t *pd;

 /* counter for testing purposes */
 count = 1;

 if ( argc == 1 ) {
        fprintf (stderr, "No arguments?\n");
        usage(argv[0]);
        exit(NO_PARAMETERS);
 }

 /* default mode in linux-server: always     */
 /* default mode in windows: reactive        */
 /* TODO default mode in DEBUG mode: passive */
 w2lanmode = "always";
 if (DEBUG) w2lanmode = "passive";
 /********************************************/

 while ( ( c = getopt( argc, argv, "m:vh" ) ) != EOF )
 {

  switch( c )
  {
   case 'm':
        w2lanmode = optarg;
	if ( strcmp(optarg, "passive") == 0 ) {
           if (DEBUG) {
              printf("DEBUG: w2lan.c: optarg passive: %s\n", optarg );
           }
        } else if ( strcmp(optarg, "reactive") == 0 ) {
           if (DEBUG) {
              printf("DEBUG: w2lan.c: optarg reactive: %s\n", optarg );
           }
        } else if ( strcmp(optarg, "always") == 0 ) {
           if (DEBUG) {
              printf("DEBUG: w2lan.c: optarg always: %s\n", optarg );
           }
        } else {
           if (DEBUG) {
              printf("DEBUG: w2lan.c: optarg other mode: %s\n", optarg );
           }
           fprintf (stderr, "Valid modes are: passive, reactive, always.\n");
           usage(argv[0]);
           exit(INVALID_MODE);
        }
        break;
   case 'v':
        display_version();
        exit(SUCCESS);
   case '?':
        if (optopt == 'm')
               fprintf (stderr, "Option -%c requires an argument.\n", optopt);
             else if (isprint (optopt))
               fprintf (stderr, "Unknown option `-%c'.\n", optopt);
             else
               fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
             usage(argv[0]);
             exit(UNKNOWN_OPTION);
   case 'h':
   default:
        usage(argv[0]);
        exit(SUCCESS);
  }
 } /* while */
 
 if (DEBUG) {
    printf("DEBUG: w2lan.c: optind = %d. argc = %d\n", optind, argc );
 }

 if ( optind != argc -1 ) {
    if ( optind < ( argc - 1 ) ) {
       fprintf (stderr, "Too many arguments.\n");
       usage(argv[0]);
       exit(TOO_MANY_ARGUMENTS);
    } else {
       fprintf (stderr, "Too little arguments.\n");
       usage(argv[0]);
       exit(TOO_LITTLE_ARGUMENTS);
    }
 }

 device = argv[ argc-1 ];

 if (DEBUG) {
    printf("DEBUG: w2lan.c: device selected: %s\n", device);
 }

 /* End of parameter processing */


 /* open the user specified device */
 if ( ( pd = pcap_open_live( device, SNAPLEN, promisc_flag, TIMEOUT_PCAP_OPEN_LIVE, errbuf ) ) == NULL ) {
    fprintf( stderr, "ERROR: pcap_open_live error: %s\n", errbuf);
    fprintf( stderr, "For more information, try from command line 'ifconfig %s'\n", device );
    exit(ERROR_OPENING_DEVICE);
 }
 /* and keep the descriptor address */
 if (DEBUG) {
     printf("DEBUG: w2lan.c: pd -> %d\n", (int)pd );
 }

 /* determine link layer type */
 if ( ( datalink = pcap_datalink( pd ) ) < 0 ) {
    fprintf( stderr, "ERROR: pcap_datalink error: %s\n", pcap_geterr(pd));
    exit(ERROR_PCAP_DATALINK); 
}

 if ( datalink != DLT_EN10MB ) {
    fprintf( stderr, "ERROR: interface is %s, not DLT_EN10MB (aka Ethernet)\n", pcap_datalink_val_to_name( datalink ) );
    exit(ERROR_DATALINK_TYPE); 
}
 /* if here, we have Ethernet */


 /* Determining if there is a wireless network underneath */
 
 /* number of wireless interfaces */
 int         interfaces = 0;

 /* temporary storage of wireless info */
 char        tmp[0x20]; /* it should be long enough */
 char        *lp;

 /* auxiliar file for reading /proc/net/wireless */
 FILE        *fdwireless;

 /* indicates user specified wireless device found*/
 int found = 0; /* not found yet */

 /* are wireless extensions present? */
 if (!(fdwireless = fopen("/proc/net/wireless", "r"))) {
   fprintf( stderr, "ERROR: no wireless extensions found!");
   fprintf( stderr, "Check if you have iwtools installed and working. Try from command line 'iwconfig %s'\n", device );
   fprintf( stderr, "If iwtools are working, check /proc. Try from command line 'mount -t proc proc /proc' and retry %s\n", argv[0] );
   exit(NO_WIRELESS_EXTENSIONS);
 }

 /* scan all the wireless adapters */

 while (fgets(tmp, 0x20, fdwireless)) {
   if ( strchr(tmp, ':') ) {
     lp = tmp + strspn(tmp, " ");
     lp[strcspn(lp, ":")] = '\0';
     /* remember if user specified wireless adapter is found */
     if ( ! strcmp( device, lp ) ) {
	found = 1;
        if (DEBUG) {
           printf("DEBUG: w2lan.c: user device selected = %s - device found = %s\n", device, lp);
        }
     } 
     interfaces++;
   }
 }

 if ( ! found ) {
   fprintf( stderr, "ERROR: device %s not found as wireless extension.\n", device );
   fprintf( stderr, "For more information about %s, try from command line 'iwconfig %s'\n", device , device );

   exit(WIRELESS_DEVICE_NOT_FOUND); 
 }
 fclose( fdwireless );

 /* if here, we have found the -user specified- underneath wireless adapter */


 /* is it configured in Ad-hoc mode ? */

 struct iwreq wrq;

 char *  OperationMode[] = { "Auto", "Ad-Hoc", "Managed", "Master", "Repeater", "Secondary" };

 /* generic raw socket descriptor */
 int opensocketretcode = -1;
 int ioctlretcode = -1;

 opensocketretcode = socket(AF_INET, SOCK_DGRAM, 0);
 if (DEBUG) {
    printf("DEBUG: w2lan.c: opensocketretcode %d (should be >=0)\n", opensocketretcode );
 }

 if ( opensocketretcode < 0) { 
    fprintf( stderr, "ERROR: could not open socket. Ret. code %d\n", opensocketretcode );
    exit(ERROR_OPENING_SOCKET); 
 } 
 
 strcpy( wrq.ifr_name, device );

 ioctlretcode = ioctl( opensocketretcode, SIOCGIWMODE, &wrq);
 if (DEBUG) {
    printf("DEBUG: w2lan.c: ioctl call: opensocketretcode: %d\n", opensocketretcode);
    printf("DEBUG: w2lan.c: ioctl call: SIOCGIWMODE: %d\n", SIOCGIWMODE);
    printf("DEBUG: w2lan.c: ioctl call: wrq: %s\n", wrq.ifr_name );
 }



 if ( ioctlretcode >= 0) {
    if (DEBUG) {
       printf("DEBUG: w2lan.c: ioctl call succesful\n");
    }
    if((wrq.u.mode < 6) && (wrq.u.mode >= 0)) {
       if (DEBUG) {
          printf("DEBUG: w2lan.c: device %s is in %s mode\n", device, OperationMode[wrq.u.mode] );
       }
       if ( wrq.u.mode != 1 ) {
          fprintf(stderr, "ERROR: device %s is in %s mode. It should be in %s mode.\n", device, OperationMode[wrq.u.mode], OperationMode[1] );
          fprintf( stderr, "To set %s from command line, try 'iwconfig %s mode %s'\n", device, device , OperationMode[1] );
          exit(INCORRECT_WIRELESS_MODE);
       } 
    } else {
       fprintf( stderr, "ERROR: Interface %s uses an unknown wireless mode\n", device );
       exit(UNKNOWN_WIRELESS_MODE); 
    }
 } else {
    fprintf(stderr, "ERROR: error calling ioctl. Ret. code %d.\n", ioctlretcode );
    exit(IOCTL_ERROR); 

 }
 close(opensocketretcode);

 /* if here, we have the wireless adapter configured in Ad-hoc mode  */

 /* obtain our mac address */

 long stat;
// int i;
 u_char addr[ETHER_ADDR_LEN];

 stat = mac_addr_sys( addr , (u_char *)device );
 memcpy( myMAC, addr , ETHER_ADDR_LEN);
 if (0 == stat) {
    if (DEBUG) {
      printf( "DEBUG: w2lan.c: myMAC: ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", myMAC[i]);
      }
      printf( "%2.2x\n", myMAC[ETHER_ADDR_LEN-1] );
    }
 } 
 else {
    fprintf( stderr, "ERROR: w2lan.c: can't get MAC address.\n");
    exit(UNABLE_GET_MY_MAC);
 }
 

 /* starting w2lan. -1 captures forever */


 /* w2lan uses 4 lists. Initialize them */
 convl = (conversationslist *)conversationslistcreate();
 countl = (counterslist *)counterslistcreate();
 pendingframesl = (pendingframeslist *)pendingframeslistcreate();
 timersl = (timerslist *)timerslistcreate();

 if (DEBUG) {
   printf("DEBUG: w2lan.c: convl: %d\n", (int)convl);
   printf("DEBUG: w2lan.c: countl: %d\n", (int)countl);
   printf("DEBUG: w2lan.c: pendingframesl: %d\n", (int)pendingframesl);
   printf("DEBUG: w2lan.c: timersl: %d\n", (int)timersl);
 }

 if (DEBUG) {
    printf("DEBUG: w2lan.c: pd -> %d\n", (int)pd );
 }

 if (DEBUG) {
    printf("DEBUG: w2lan.c: w2lan mode: %s\n", w2lanmode );
 }

 int retvalue;
 int selectretvalue;
 fd_set set_read;
 

 // Initialize the timer

 timerrequestglobal = (struct timeval *)malloc(sizeof(struct timeval));
 timerrequest = NULL;

 // To generate random numbers. Use of time to create seed.
// struct tm t;
// time(&t);
// u_int16_t newseed[3];
 u_int16_t newseed[3] = { 1 , 3 , 5 };
 //newseed[1] = t.tm_sec;
 //newseed[2] = t.tm_min;
 //newseed[3] = t.tm_hour;

// if (DEBUG) {
//    printf("DEBUG: w2lan.c: t.tm_sec = %d; t.tm_min = %d; t.tm_hour = %d; \n", t.tm_sec , t.tm_min //, t.tm_hour );
// }
 
 seed48(newseed);
 if (DEBUG) {
    printf("DEBUG: w2lan.c: newseed -> newseed[0] = %x; newseed[1] = %x; newseed[2] = %x\n", newseed[0], newseed[1], newseed[2]  );
 }

 // Default starting value in reactive mode
 counteractivewindow = 0;
 activewindow = 1; // starting "speaking" w2lan


 int fd = pcap_get_selectable_fd( pd );
 if (DEBUG) {
    printf("DEBUG: w2lan.c: fd (= %d) = pcap_get_selectable_fd( pd (= %d))\n", (int)fd , (int)pd );
 }

 // Permanent regime first line //////////////////////////////////////////
 do {

    // Initializing file descriptor set
    FD_ZERO( &set_read );
    FD_SET ( fd , &set_read );

    if (DEBUG) {
       if (timerrequest != NULL) {
          printf("DEBUG: w2lan.c: timerrequest->tv_sec  -> %d\n", (int)(timerrequest->tv_sec) );
          printf("DEBUG: w2lan.c: timerrequest->tv_usec -> %d\n", (int)(timerrequest->tv_usec) );
       } else {
          printf("DEBUG: w2lan.c: timerrequest is NULL (never expires)\n");
       }
    }

//    if (DEBUG) {
//       printf("\nDEBUG: w2lan.c: FD_SETSIZE = %d\n", FD_SETSIZE );
//    }

    if (DEBUG) {
       printf("\nDEBUG: w2lan.c: entering select...\n");
    }
//    selectretvalue = select( FD_SETSIZE, &set_read , NULL, NULL , timerrequest ); 
    selectretvalue = select( fd + 1 , &set_read , NULL, NULL , timerrequest ); 

    if ( selectretvalue == 1 ) {

      // the fourth field is to indicate the network socket to the callback function
      // many people say "useless field", but it is useful i.e. for the callback function
      // to send packets in the same socket we opened to receive packets
      if (DEBUG) {
         printf("DEBUG: w2lan.c: selectretvalue = %d. calling pcap_loop...\n", selectretvalue );
      }

      retvalue = pcap_loop( pd , 1 , w2lan_callback, (unsigned char *)pd );

    } else if ( selectretvalue == 0 ) {

      if (DEBUG) {
         printf("DEBUG: w2lan.c: selectretvalue = %d. timerrequest has expired.\n", selectretvalue );
      }
      // timerrequest has expired
// TODO comentar seg. linia
//timerrequest = NULL;
// TODO descomentar seg.linia
      timer_expired( pd );

    } else {

       // That should not happen :-(
       if (DEBUG) {
         printf("DEBUG: w2lan.c: Error in select function.\n");
         printf("DEBUG: w2lan.c: select returned: selectretvalue = %d\n", selectretvalue );
       }
       fprintf( stderr, "ERROR: w2lan.c: 'select' function returned %d.\n", selectretvalue );
       exit(NEVER_HERE);

    }

 } while (1);
 // Permanent regime last line ///////////////////////////////////////////



 if (DEBUG) {
   printf("DEBUG: w2lan.c: End of program. Stable version should never arrive here.\n");
   printf("DEBUG: w2lan.c: pcap_loop retvalue = %d\n", retvalue );
 }

 exit(NEVER_HERE);

}  /* main */



long mac_addr_sys ( u_int8_t *addr, unsigned char * dvce ) {

    struct ifreq ifr;
    struct ifreq *IFR;
    struct ifconf ifc;
    char buf[1024];
    int s, i;
    int ok = 0;

    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s==-1) {
        return -1;
    }

    ifc.ifc_len = sizeof(buf);
    ifc.ifc_buf = buf;
    ioctl(s, SIOCGIFCONF, &ifc);

    IFR = ifc.ifc_req;
    for (i = ifc.ifc_len / sizeof(struct ifreq); --i >= 0; IFR++) {

        strcpy(ifr.ifr_name, (char *)dvce );
        if (ioctl(s, SIOCGIFFLAGS, &ifr) == 0) {
            if (! (ifr.ifr_flags & IFF_LOOPBACK)) {
                if (ioctl(s, SIOCGIFHWADDR, &ifr) == 0) {
                    ok = 1;
                    break;
                }
            }
        }
    }

    close(s);
    if (ok) {
        memcpy( addr, ifr.ifr_hwaddr.sa_data, ETHER_ADDR_LEN );
    }
    else {
        return -1;
    }
    return 0;
}

