/***************************************************************************
 *   Copyright (C) 2007 by Francesc Burrull i Mestres.                     *
 *   francesc.burrull@upct.es                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _W2LAN_CORE_H_
#include "w2lan_core.h"
#endif


//void announce_received( char* payload , u_int8_t *fromMAC, unsigned char *sd) {
void announce_received( unsigned char *extra , const struct pcap_pkthdr *pkthdr , const unsigned char *packet ) {


//   if (DEBUG) {
//      printf( "DEBUG: w2lan_core.c: test ownMAC: [ ");
//      int i;
//      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
//        printf("%2.2x:", ownMAC[i]);
//      }
//      printf("%2.2x ]\n", ownMAC[ETHER_ADDR_LEN-1]);
//   }


   /* packet capture descriptor */
   pcap_t *pd = (pcap_t *)extra;

   /* cast to ethernet header */
   struct ether_header *eptr;
   eptr = (struct ether_header *) packet;

   int pcap_sendpacket_retcode;

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: announce_received(): pkthdr->len = %d. ETH_P_W2LAN_ANNOUNCE_LEN = %d\n", pkthdr->len , ETH_P_W2LAN_ANNOUNCE_LEN );
   }


   /* if malformed return */
   if ( pkthdr->len != ETH_P_W2LAN_ANNOUNCE_LEN ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: announce_received(): malformed announce packet (size = %d bytes).\n", pkthdr->len );
      }
      return;
   }

   /* if not broadcast return */
   if ( memcmp( eptr->ether_dhost, ether_aton( "FF:FF:FF:FF:FF:FF" ), ETHER_ADDR_LEN ) != 0 ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: announce_received(): malformed non-broadcast w2lan announce (size = %d bytes).\n", pkthdr->len );
      }
      return;
   }

   /* here we have a valid w2lan announce */

   // check if conversation is in conversationslist return else ...(lots)
   /* packet+28 is the position in packet where conversationID appears */
   if ( conversationslistsearch( (u_int8_t *)(packet+28) , (conversationslist)convl ) == NULL ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: announce_received(): packet not in the list.\n");
      }
      // not in the list. add to the conversationslist

      u_int16_t ethernettypeaux;
      memcpy( (&ethernettypeaux), packet+26, 2);
      conversationslistadd( (u_int8_t *)(packet+28) , (u_int8_t *)(packet+14), (u_int8_t *)(packet+20), ntohs( ethernettypeaux ) , (conversationslist *)&convl );

   } else {
      // in the list, do nothing else
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: announce_received(): packet in the list.\n");
      }
      return;
   }

   unsigned char *w2lanreqpacket;

//   int bytesallocated;
//
// FBM: I don't know yet wich value I will use....
//   if ( ETH_P_W2LAN_ANNOUNCE_LEN < ETH_ZLEN) {
//      bytesallocated = ETH_ZLEN;
//   }
//   else {
//      bytesallocated = ETH_P_W2LAN_ANNOUNCE_LEN;
//   } 

   w2lanreqpacket = (unsigned char *)malloc(ETH_P_W2LAN_ANNOUNCE_LEN*sizeof(unsigned char));
   assert(w2lanreqpacket != NULL);
   /* Initialize reqpacket */
   memcpy( w2lanreqpacket, "", 1);
//   char strannounce[2] = "";
//   strannounce[1] = htons(0x6901);
//   strannounce[2] = hton(0x01);

//temporal
//  
//////////

//   if (DEBUG) printf("DEBUG: w2lan_core.c: [ %s ] (%d bytes)\n", w2lanreqpacket, strlen(w2lanreqpacket) );

   memcpy(  w2lanreqpacket , eptr->ether_shost , ETHER_ADDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: announce_received(): w2lanreqpacket dst::::::::::::::: [ ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanreqpacket[i]);
      }
      printf("%2.2x ]\n", w2lanreqpacket[ETHER_ADDR_LEN-1]);
   }


   memcpy( w2lanreqpacket + ETHER_ADDR_LEN , myMAC , ETHER_ADDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: announce_received(): w2lanreqpacket src::::::::::::::: [ ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanreqpacket[i+6]);
      }
      printf("%2.2x ]\n", w2lanreqpacket[ETHER_ADDR_LEN-1+6]);
   }

   u_int16_t reqtype = htons( ETH_P_W2LAN_REQUEST );
   memcpy( w2lanreqpacket+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (unsigned char *)(&reqtype) , 2 );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: announce_received(): w2lanreqpacket type:::::::::::::: [ ");
      printf("0x%2.2X%2.2X ]\n", w2lanreqpacket[12] , w2lanreqpacket[13]);
   }

   memcpy( w2lanreqpacket+ETHER_ADDR_LEN+ETHER_ADDR_LEN+2 , packet+28 , ETH_P_W2LAN_CONVERSATIONID_LEN );
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: announce_received(): w2lanreqpacket payload (%d bytes): [ ", ETH_P_W2LAN_CONVERSATIONID_LEN );
      int i;
      for (i=0; i< ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanreqpacket[i+ETHER_HDR_LEN]);
      }
      printf("%2.2x ]\n", w2lanreqpacket[ETHER_HDR_LEN+ETHER_ADDR_LEN-1]);
   }


   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: announce_received(): pd -> %d\n", (int)pd );
   }


   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: announce_received(): packet to sendpacket (%d bytes): [ ", ETH_P_W2LAN_REQUEST_LEN);
      int i;
      for (i=0; i<ETH_P_W2LAN_REQUEST_LEN-1; ++i) {
        printf("%2.2x:", w2lanreqpacket[i]);
      }
      printf( "%2.2x ]\n", w2lanreqpacket[ETH_P_W2LAN_REQUEST_LEN-1] );
   }




   if ( ( pcap_sendpacket_retcode = pcap_sendpacket( pd , w2lanreqpacket , ETH_P_W2LAN_REQUEST_LEN ) ) < 0 )
   {
      fprintf(stderr,"\nError sending the packet: %s\n", pcap_geterr( pd ));
   }

   free( w2lanreqpacket );

   if (DEBUG) {
//      printf("DEBUG: w2lan_core.c: pcap_sendpacket retcode: %d (should be 0)\n", retcode );
      printf("DEBUG: w2lan_core.c: announce_received(): pcap_sendpacket pcap_sendpacket_retcode: %d (should be >= 0)\n", pcap_sendpacket_retcode );
   }

}


void request_received( unsigned char *extra , const struct pcap_pkthdr *pkthdr , const unsigned char *packet ) {

   /* packet capture descriptor */
//   pcap_t *pd = (pcap_t *)extra;

   /* cast to ethernet header */
   struct ether_header *eptr;
   eptr = (struct ether_header *) packet;

//   int pcap_sendpacket_retcode;

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: request_received(): pkthdr->len = %d. ETH_P_W2LAN_REQUEST_LEN = %d\n", pkthdr->len , ETH_P_W2LAN_REQUEST_LEN );
   }


   /* if malformed return */
   if ( pkthdr->len != ETH_P_W2LAN_REQUEST_LEN ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: request_received(): malformed request packet (size = %d bytes).\n", pkthdr->len );
      }
      return;
   }

   /* here we have a valid w2lan request */

   if ( counterslistsearch( (u_int8_t *)(packet+14) , (counterslist)countl ) == NULL ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: request_received(): packet not in the counters list.\n");
      }
      // not in the list. do nothing
      return;
   } else {
// TODO test/debug the counterslist
      // in the list. increase counter associated with conversationid
      counterslistincrease( (u_int8_t *)(packet+14) , (counterslist)countl );
   }

}


void data_received( unsigned char *extra , const struct pcap_pkthdr *pkthdr , const unsigned char *packet ) {

   /* packet capture descriptor */
   pcap_t *pd = (pcap_t *)extra;

   /* cast to ethernet header */
   struct ether_header *eptr;
   eptr = (struct ether_header *) packet;

   int pcap_sendpacket_retcode;


   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: data_received(): pkthdr->len = %d (should be %d >=  %d  >= %d)\n", (pkthdr->len) , ETH_P_W2LAN_DATA_MIN_LEN , (pkthdr->len) , ETH_P_W2LAN_DATA_MAX_LEN );
   }

   if ( ( (pkthdr->len) >= ETH_P_W2LAN_DATA_MAX_LEN ) ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: data_received(): malformed request packet (size = %d bytes).\n", pkthdr->len );
      }
      return;
   }

   /* if not broadcast return */
   if ( ( memcmp( eptr->ether_dhost, (u_int8_t *)ether_aton( "FF:FF:FF:FF:FF:FF" ), ETHER_ADDR_LEN ) ) != 0 ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: data_received(): malformed non-broadcast w2lan data (size = %d bytes).\n", pkthdr->len );
      }
      return;
   }

   /* here we have a w2lan data packet.  */


   // see if conversation is pending...
   /* packet+6 is the position in packet where conversationID appears */
   conversationslist conv = conversationslistsearch( (u_int8_t *)(packet+6) , (conversationslist)convl );

   if ( conv == NULL ) {
      // conversation  not in list. do nothing

      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: data_received(): packet not in the conversationslist convl (do nothing).\n");
      }
      return;
   } 
   else {
      // see if the conversation is pending
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: data_received(): packet in the conversationslist convl.\n");
      }
 
      if ( conversationpending( conv ) ) {
         // conversation is pending. we have to clear its pending field

         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): packet in the list with conversation pending.\n");
         }
         conversationclearpending( conv );

         if (DEBUG) {
            if ( conversationpending( conv ) ) {
               printf("DEBUG: w2lan_core.c: data_received(): ERROR clearing conversation pending.\n");
            } else  {
               printf("DEBUG: w2lan_core.c: data_received(): conversation pending cleared.\n");
            }
         }

         // conversation was pending. Is it ours (myMAC, broadcast, TODO multicast) ? ask Josemaria
         // multicast if first bit in network format is 1
         char multicast;
         multicast = conversationgetmacdst( conv )[0];
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): first byte mac dst: %X\n", multicast);
         }
         multicast = multicast & 0x01;
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): multicast bit: %X\n", multicast);
         }
 
         if ( ( memcmp( conversationgetmacdst( conv ), ether_aton( "FF:FF:FF:FF:FF:FF" ), ETHER_ADDR_LEN ) == 0 ) || ( memcmp( conversationgetmacdst( conv ), myMAC , ETHER_ADDR_LEN ) == 0) || ( multicast == 1 ) ) {

            // the frame is ours.  we have to reconstruct the ethernet frame and sendpacket it
            if (DEBUG) {
               printf( "DEBUG: w2lan_core.c: data_received(): conversationgetmacdst( conv ) [ ");
               int i;
               u_int8_t kk[6]="";
               memcpy( kk, conversationgetmacdst( conv ), 6);
               for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
                 printf("%2.2x:", kk[i]);
               }
               printf("%2.2x ]\n", kk[ETHER_ADDR_LEN-1]);
            }

            if (DEBUG) {
               printf("DEBUG: w2lan_core.c: data_received(): Reconstructing ethernet frame...\n");
            }

            unsigned char *ethernetframe;

            ethernetframe = (unsigned char *)malloc(ETHER_MAX_LEN*sizeof(unsigned char));
            assert(ethernetframe != NULL);
            /* Initialize datapacket */
            memcpy( ethernetframe, "", 1);

            memcpy( ethernetframe , conversationgetmacdst( conv ) , ETHER_ADDR_LEN );

            if (DEBUG) {
               printf( "DEBUG: w2lan_core.c: data_received(): reconstructed ethernet dst:::::::::: [ ");
               int i;
               for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
                 printf("%2.2x:", ethernetframe[i]);
               }
               printf("%2.2x ]\n", ethernetframe[ETHER_ADDR_LEN-1]);
            }

            memcpy( ethernetframe + ETHER_ADDR_LEN , conversationgetmacsrc( conv ) , ETHER_ADDR_LEN );

            if (DEBUG) {
               printf( "DEBUG: w2lan_core.c: data_received(): reconstructed ethernet src:::::::::::: [ ");
               int i;
               for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
                 printf("%2.2x:", ethernetframe[i+ETHER_ADDR_LEN]);
               }
               printf("%2.2x ]\n", ethernetframe[ETHER_ADDR_LEN-1+ETHER_ADDR_LEN]);
            }

            u_int16_t ethernettype = htons( conversationgetmactype( conv ) );
            memcpy( ethernetframe+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (u_int8_t *)(&ethernettype) , 2 );

            if (DEBUG) {
               printf( "DEBUG: w2lan_core.c: data_received(): reconstructed ethernet type:::::::::::::: [ ");
               printf("0x%2.2X%2.2X ]\n", ethernetframe[12] , ethernetframe[13]);
            }

            memcpy( ethernetframe + ETHER_HDR_LEN , packet + ETHER_HDR_LEN , (pkthdr->len)-ETHER_HDR_LEN );

            if (DEBUG) {
               printf( "DEBUG: w2lan_core.c: data_received(): reconstructed ethernet payload::::::::: [ ");
               int i;
               for (i=0; i<(pkthdr->len)-ETHER_HDR_LEN-1; ++i) {
                 printf("%2.2x:", ethernetframe[i+ETHER_HDR_LEN]);
               }
               printf("%2.2x ]\n", ethernetframe[(pkthdr->len)-1]);
            }

            if (DEBUG) {
               printf( "DEBUG: w2lan_core.c: data_received(): ethernet frame to sendpacket (%d bytes): [ ", pkthdr->len );
               int i;
               for (i=0; i<(pkthdr->len)-1; ++i) {
                 printf("%2.2x:", ethernetframe[i]);
               }
               printf( "%2.2x ]\n", ethernetframe[(pkthdr->len)-1] );
            }

            if ( ( pcap_sendpacket_retcode = pcap_sendpacket( pd , ethernetframe , pkthdr->len ) ) < 0 ) {
               fprintf(stderr,"\nError sending the ethernet frame: %s\n", pcap_geterr( pd ));
            }

            free( ethernetframe );

            if (DEBUG) {
               printf("DEBUG: w2lan_core.c: data_received(): pcap_sendpacket pcap_sendpacket_retcode: %d (should be >= 0)\n", pcap_sendpacket_retcode );
            }

         } else {
            // ethernet frame not ours. do nothing 
            if (DEBUG) {
               printf("DEBUG: w2lan_core.c: data_received(): Ethernet frame is not ours. do nothing.\n");
            }
         }


         // conversation was pending. we prepare next announce adding ethernet payload to pending list, a new counter to counterslist, and a new timer to timerslist

// TODO check the lists
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Entering pendingframeslistput..[ ");
            int i;
            for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", (packet+6)[i]);
            }
            printf("%2.2x ]\n", (packet+6)[ETHER_ADDR_LEN-1]);
         }
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Entering pendingframeslistput..[ ");
            int i;
            for (i=0; i<(pkthdr->len)-ETHER_HDR_LEN-1; ++i) {
              printf("%2.2x:", (packet+14)[i]);
            }
            printf("%2.2x ]\n", (packet+14)[(pkthdr->len)-ETHER_HDR_LEN-1]);

         }
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Entering pendingframeslistput: (pkthdr->len)-ETHER_HDR_LEN: %d\n", (pkthdr->len)-ETHER_HDR_LEN);
         }
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Entering pendingframeslistput: pendingframesl: %d\n", (int)pendingframesl);
         }
         pendingframeslistput( (u_int8_t*)(packet+6) , (u_int8_t*)(packet+14) , (pkthdr->len)-ETHER_HDR_LEN , (pendingframeslist *)&pendingframesl );
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Exiting pendingframeslistput. (pkthdr->len)-ETHER_HDR_LEN = %d\n", (pkthdr->len)-ETHER_HDR_LEN);
         }


         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Entering counterslistintoaslast..[ ");
            int i;
            for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", (packet+6)[i]);
            }
            printf("%2.2x ]\n", (packet+6)[ETHER_ADDR_LEN-1]);
         }
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Entering counterslistintoaslast: countl: %d\n", (int)countl);
         }
         counterslistintoaslast( (u_int8_t*)(packet+6) , (counterslist *)&countl );
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Exiting counterslistintoaslast...\n");
         }

         // TODO Timers... (not TODO anymore ?)

         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Timers, here we go...\n");
         }
/////// TODO//////////
         struct timeval t1;
         int gettimeofdayretcode = gettimeofday( &t1 , NULL );
         if ( gettimeofdayretcode == 0 ) {
            if (DEBUG) {
               printf("DEBUG: w2lan_core.c: data_received(): gettimeofday: %d secs, %d usecs.\n", (int)(t1.tv_sec) , (int)(t1.tv_usec) );
            }
         } else {
            fprintf( stderr, "ERROR: gettimeofday failed (ret. code %d).\n", gettimeofdayretcode );
            exit(GET_TIME_OF_THE_DAY_ERROR);
         }
////////////////

         // prepare the local timer
         struct timeval aux;
         aux.tv_sec = TIMEOUT_REQUEST_SECONDS;
         aux.tv_usec = TIMEOUT_REQUEST_USECONDS;
         timeval_add( &t1, &t1, &aux );

         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): entering timerslistempty...\n");
         }

         if ( timerslistempty( timersl ) == 1 ) {
            // list empty. set the global timer
            if (DEBUG) {
               printf("DEBUG: w2lan_core.c: data_received(): updating timerrequest...\n");
            }
            timerrequest = timerrequestglobal;
            timerrequest -> tv_sec = TIMEOUT_REQUEST_SECONDS;
            timerrequest -> tv_usec = TIMEOUT_REQUEST_USECONDS;
         }

         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): entering timerslistintoaslast...\n");
         }

         timerslistintoaslast( (u_int8_t*)(packet+6) , &t1 , (timerslist *)&timersl );
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): Timers, done ! \n");
         }
////////////////////
         // after updating lists,  we reannounce the ethernet frame.
         unsigned char *w2lanannouncepacket;

         w2lanannouncepacket = (unsigned char *)malloc(ETH_P_W2LAN_ANNOUNCE_LEN*sizeof(unsigned char));
         assert(w2lanannouncepacket != NULL);

         /* Initialize announce packet */
         memcpy( w2lanannouncepacket, "", 1);

         memcpy( w2lanannouncepacket , (u_int8_t *)ether_aton("FF:FF:FF:FF:FF:FF") , ETHER_ADDR_LEN );
         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket dst:::::::::: [ ");
            int i;
            for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", w2lanannouncepacket[i]);
            }
            printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN-1]);
         }

         memcpy( w2lanannouncepacket + ETHER_ADDR_LEN , myMAC , ETHER_ADDR_LEN );
         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket src:::::::::::: [ ");
            int i;
            for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", w2lanannouncepacket[i+ETHER_ADDR_LEN]);
            }
            printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN-1+ETHER_ADDR_LEN]);
         }

         u_int16_t announcetype = htons( ETH_P_W2LAN_ANNOUNCE );
         memcpy( w2lanannouncepacket+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (u_int8_t *)(&announcetype) , 2 );
         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket type:::::::::::::: [ ");
            printf("0x%2.2X%2.2X ]\n", w2lanannouncepacket[12] , w2lanannouncepacket[13]);
         }

         memcpy( w2lanannouncepacket + ETHER_HDR_LEN , (u_int8_t *)conversationgetmacdst(conv) , ETHER_ADDR_LEN );
         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket src:::::::::::: [ ");
            int i;
            for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", w2lanannouncepacket[i+ETHER_HDR_LEN]);
            }
            printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN-1+ETHER_HDR_LEN]);
         }

         memcpy( w2lanannouncepacket + ETHER_HDR_LEN + ETHER_ADDR_LEN, (u_int8_t *)conversationgetmacsrc(conv) , ETHER_ADDR_LEN );
         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket src:::::::::::: [ ");
            int i;
            for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", w2lanannouncepacket[i+ETHER_HDR_LEN+ETHER_ADDR_LEN]);
            }
            printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN+ETHER_ADDR_LEN-1+ETHER_HDR_LEN]);
         }

         u_int16_t storedethernettype = htons( conversationgetmactype( conv ) );
         memcpy( w2lanannouncepacket+ETHER_HDR_LEN+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (u_int8_t *)(&storedethernettype) , 2 );
         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket type:::::::::::::: [ ");
            printf("0x%2.2X%2.2X ]\n", w2lanannouncepacket[ETHER_HDR_LEN+12] , w2lanannouncepacket[ETHER_HDR_LEN+13]);
         }

         /* packet+6 is the position in packet where conversationID appears */
         memcpy( w2lanannouncepacket+ETHER_HDR_LEN+ETHER_ADDR_LEN+ETHER_ADDR_LEN+2 , packet+6 , ETH_P_W2LAN_CONVERSATIONID_LEN );
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): w2lanannouncepacket payload (%d bytes): [ ", ETH_P_W2LAN_CONVERSATIONID_LEN );
            int i;
            for (i=0; i< ETHER_ADDR_LEN-1; ++i) {
              printf("%2.2x:", w2lanannouncepacket[i+ETHER_HDR_LEN+ETHER_HDR_LEN]);
            }
            printf("%2.2x ]\n", w2lanannouncepacket[ETHER_HDR_LEN+ETHER_HDR_LEN+ETHER_ADDR_LEN-1]);
         }


         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): pd -> %d\n", (int)pd );
         }


         if (DEBUG) {
            printf( "DEBUG: w2lan_core.c: data_received(): packet to sendpacket (%d bytes): [ ", ETH_P_W2LAN_ANNOUNCE_LEN);
            int i;
            for (i=0; i<ETH_P_W2LAN_ANNOUNCE_LEN-1; ++i) {
               printf("%2.2x:", w2lanannouncepacket[i]);
            }
            printf( "%2.2x ]\n", w2lanannouncepacket[ETH_P_W2LAN_ANNOUNCE_LEN-1] );
         }

         if ( ( pcap_sendpacket_retcode = pcap_sendpacket( pd , w2lanannouncepacket , ETH_P_W2LAN_ANNOUNCE_LEN ) ) < 0 ) { 
            fprintf(stderr,"\nError sending the packet: %s\n", pcap_geterr( pd ));
         }

         free( w2lanannouncepacket );

         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): pcap_sendpacket pcap_sendpacket_retcode: %d (should be >= 0)\n", pcap_sendpacket_retcode );
         }

      } else {
         // conversation not pending. do nothing

         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: data_received(): packet in the list with conversation not pending (do nothing).\n");
         }

         return;
      }  // else - conversation not pending
   }  // else - not in conversationslist

}  // datareceived()


void timer_expired(pcap_t *pd) {

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): Starting timer_expired()...\n");
   }

   // Timers...

   struct timeval t;
   struct timeval t1;
   struct timeval t2;
   u_int8_t conversationID[ETH_P_W2LAN_CONVERSATIONID_LEN];

//   conversationID = (u_int8_t *)malloc(ETH_P_W2LAN_CONVERSATIONID_LEN*sizeof(u_int8_t));
//   assert(conversationID != NULL);


   // extract expired timer
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): entering timerslistfirstinqueue...\n");
   }
   timerslistfirstinqueue( conversationID , &t1, (timerslist *)&timersl );
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): timerslistfirstinqueue done. conversationID: [ ");
      int i;
      for (i=0; i<ETH_P_W2LAN_CONVERSATIONID_LEN-1; ++i) {
        printf("%2.2x:", (conversationID)[i]);
      }
      printf("%2.2x ]\n", (conversationID)[ETH_P_W2LAN_CONVERSATIONID_LEN-1]);
   }



   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): entering timerslistempty...\n");
   }
   if ( timerslistempty( timersl ) == 0 ) {
      // timersl not empty. set the global timer
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: timer_expired(): updating timerrequest from timersl..\n");
      }

      timerslistsuccessor( conversationID , &t1 , timersl );

      int gettimeofdayretcode = gettimeofday( &t2 , NULL );
      if ( gettimeofdayretcode == 0 ) {
         if (DEBUG) {
            printf("DEBUG: w2lan_core.c: timer_expired(): gettimeofday: %d secs, %d usecs.\n", (int)(t2.tv_sec) , (int)(t2.tv_usec) );
         }
      } else {
         fprintf( stderr, "ERROR: gettimeofday failed (ret. code %d).\n", gettimeofdayretcode );
         exit(GET_TIME_OF_THE_DAY_ERROR);
      }

      if ( ( timeval_subtract ( &t , &t1 , &t2 ) ) != 0 ) {
         fprintf( stderr, "ERROR: timeval_subtract returned a negative time.\n");
         exit(TIMEVAL_SUBSTRACT_ERROR);
      }

      if ( timerrequest == NULL ) {
         timerrequest = timerrequestglobal;
      }
      timerrequest -> tv_sec = t.tv_sec;
      timerrequest -> tv_usec = t.tv_usec;
   } else {
      // timersl empty. disable the global timer
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: timer_expired(): timersl empty.\n");
      }
      timerrequest = NULL;
   }

   // counters. Here we also obtain conversationID

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): entering counterslistfirstinqueue (countl = %d)\n", (int)countl);
   }
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): Entering counterslistfirstinqueue..[ ");
      int i;
      for (i=0; i<ETH_P_W2LAN_CONVERSATIONID_LEN-1; ++i) {
        printf("%2.2x:", (conversationID)[i]);
      }
      printf("%2.2x ]\n", (conversationID)[ETH_P_W2LAN_CONVERSATIONID_LEN-1]);
   }

   int numrequests;
   numrequests = counterslistfirstinqueue( conversationID , (counterslist *)&countl );

// TODO trampa eliminar seguent linia TODO TODO TODO
//numrequests = 1;
///TODO TODO TODO TODO TODO TODO TODO/////////////////////////////

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): numrequests = %d.\n", numrequests );
   }

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): counterslistfirstinqueue done.\n");
   }


   // pending frames lists...

   int pcap_sendpacket_retcode;
   u_int16_t w2landatapacketsize;
   unsigned char *w2landatapacket;
   w2landatapacket = (unsigned char *)malloc(ETHER_MAX_LEN*sizeof(unsigned char));
   assert(w2landatapacket != NULL);
   /* Initialize datapacket */
   memcpy( w2landatapacket, "", 1);

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): Entering pendingframeslistextract..[ ");
      int i;
      for (i=0; i<ETH_P_W2LAN_CONVERSATIONID_LEN-1; ++i) {
        printf("%2.2x:", (conversationID)[i]);
      }
      printf("%2.2x ]\n", (conversationID)[ETH_P_W2LAN_CONVERSATIONID_LEN-1]);
   }
   pendingframeslistextract( conversationID , w2landatapacket+ETHER_HDR_LEN , &w2landatapacketsize, (pendingframeslist *)&pendingframesl );

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): w2landatapacketsize = %d.\n", (u_int16_t)w2landatapacketsize );
   }
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): w2landatapacket ~ (%d bytes): [ ", w2landatapacketsize );
      int i;
      for (i=0; i< w2landatapacketsize-1; ++i) {
        printf("%2.2x:", (w2landatapacket+ETHER_HDR_LEN)[i]);
      }
      printf("%2.2x ]\n", w2landatapacket[w2landatapacketsize+ETHER_HDR_LEN-1]);
   }


   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: timer_expired(): numrequests = %d.\n", numrequests );
   }
   if ( numrequests > 0 ) {

      memcpy( w2landatapacket , ether_aton( "FF:FF:FF:FF:FF:FF" ) , ETHER_ADDR_LEN );
      if (DEBUG) {
         printf( "DEBUG: w2lan_core.c: timer_expired(): w2landatapacket dst:::::::::::: [ ");
         int i;
         for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
           printf("%2.2x:", w2landatapacket[i]);
         }
         printf("%2.2x ]\n", w2landatapacket[ETHER_ADDR_LEN-1]);
      }

      memcpy( w2landatapacket+ETHER_ADDR_LEN , conversationID , ETHER_ADDR_LEN );
      if (DEBUG) {
         printf( "DEBUG: w2lan_core.c: timer_expired(): w2landatapacket src:::::::::::: [ ");
         int i;
         for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
           printf("%2.2x:", w2landatapacket[i+ETHER_ADDR_LEN]);
         }
         printf("%2.2x ]\n", w2landatapacket[ETHER_ADDR_LEN-1+ETHER_ADDR_LEN]);
      }

      u_int16_t datatype = htons( ETH_P_W2LAN_DATA );
      memcpy( w2landatapacket+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (unsigned char *)(&datatype) , 2 );
      if (DEBUG) {
         printf( "DEBUG: w2lan_core.c: timer_expired(): w2landatapacket type:::::::::::::: [ ");
         printf("0x%2.2X%2.2X ]\n", w2landatapacket[12] , w2landatapacket[13]);
      }

      // check ethernet data field
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: timer_expired(): w2landatapacket payload (%d bytes): [ ", w2landatapacketsize );
         int i;
         for (i=0; i< w2landatapacketsize-1; ++i) {
           printf("%2.2x:", w2landatapacket[i+ETHER_HDR_LEN]);
         }
         printf("%2.2x ]\n", w2landatapacket[w2landatapacketsize+ETHER_HDR_LEN-1]);
      }


      if (DEBUG) {
         printf( "DEBUG: w2lan_core.c: timer_expired(): packet to sendpacket (%d bytes): [ ", w2landatapacketsize+ETHER_HDR_LEN);
         int i;
         for (i=0; i<w2landatapacketsize+ETHER_HDR_LEN-1; ++i) {
            printf("%2.2x:", w2landatapacket[i]);
         }
         printf( "%2.2x ]\n", w2landatapacket[w2landatapacketsize+ETHER_HDR_LEN-1] );
      }

      if ( ( pcap_sendpacket_retcode = pcap_sendpacket( pd , w2landatapacket , w2landatapacketsize+ETHER_HDR_LEN ) ) < 0 ) { 
         fprintf(stderr,"\nError sending the packet: %s\n", pcap_geterr( pd ));
      }


   }

   free(w2landatapacket);
   return;

}

/////////////////////////////////
/////////////////////////////////
/////////////////////////////////

void generate_announce( unsigned char *extra , const struct pcap_pkthdr *pkthdr , const unsigned char *packet ) {

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Starting generate_announce()...\n");
   }

   /* packet capture descriptor */
   pcap_t *pd = (pcap_t *)extra;

   /* cast to ethernet header */
   struct ether_header *eptr;
   eptr = (struct ether_header *) packet;

   int pcap_sendpacket_retcode;


   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): pkthdr->len = %d (should be %d >=  %d  >= %d)\n", (pkthdr->len) , ETHER_MIN_LEN , (pkthdr->len) , ETH_FRAME_LEN );
   }

   unsigned char *ethernetframe;

   ethernetframe = (unsigned char *)malloc(ETHER_MAX_LEN*sizeof(unsigned char));
   assert(ethernetframe != NULL);
   /* Initialize datapacket */
   memcpy( ethernetframe, "", 1);

   memcpy( ethernetframe , packet+ETHER_HDR_LEN , (pkthdr->len)-ETHER_HDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): ethernet payload::: [ ");
      int i;
      for (i=0; i<(pkthdr->len)-ETHER_HDR_LEN-1; ++i) {
        printf("%2.2x:", ethernetframe[i]);
      }
      printf("%2.2x ]\n", ethernetframe[(pkthdr->len)-ETHER_HDR_LEN-1]);
   }

   // Random numbers are XX XX XX XX XX XX
   long longconversationID = lrand48(); // for the 4 least significative bytes
   long longconversationIDhigh = lrand48(); // for the 2 most significative bytes
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): longconversationID = %ld. sizeof(longconversationID) = %d\n", longconversationID , sizeof(longconversationID) );
      printf("DEBUG: w2lan_core.c: generate_announce(): longconversationIDhigh = %ld. sizeof(longconversationID) = %d\n", longconversationIDhigh , sizeof(longconversationIDhigh) );
   }
   u_int8_t conversationID[ETH_P_W2LAN_CONVERSATIONID_LEN] = { 0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF };
   memcpy( conversationID , (u_int8_t *)(& longconversationIDhigh), ETH_P_W2LAN_CONVERSATIONID_LEN );
   memcpy( conversationID+2 , (u_int8_t *)(& longconversationID), ETH_P_W2LAN_CONVERSATIONID_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): conversationID: [ ");
      int i;
      for (i=0; i<ETH_P_W2LAN_CONVERSATIONID_LEN-1; ++i) {
        printf("%2.2x:", conversationID[i]);
      }
      printf("%2.2x ]\n", conversationID[ETH_P_W2LAN_CONVERSATIONID_LEN-1]);
   }

   // Add conversationID to the conversationslist convl as pending
   u_int16_t ethernettypeaux;
   memcpy( (&ethernettypeaux), packet+12, 2);
   conversationslistadd( conversationID , (u_int8_t *)(packet) , (u_int8_t *)(packet+6) , ntohs(ethernettypeaux) , (conversationslist *)&convl );


   // Add conversationID+frame to the pending list pendingframesl
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Entering pendingframeslistput: pendingframesl: %d\n", (int)pendingframesl);
   }
   pendingframeslistput( conversationID , (u_int8_t*)(ethernetframe) , (pkthdr->len)-ETHER_HDR_LEN , (pendingframeslist *)&pendingframesl );
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Exiting pendingframeslistput. (pkthdr->len)-ETHER_HDR_LEN = %d\n", (pkthdr->len)-ETHER_HDR_LEN);
   }

   // Add new counter to countl
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Entering counterslistintoaslast: countl: %d\n", (int)countl);
   }
   counterslistintoaslast( conversationID , (counterslist *)&countl );
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Exiting counterslistintoaslast...\n");
   }

   // Add new timer to timersl
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Timers, here we go...\n");
   }

   struct timeval t1;
   int gettimeofdayretcode = gettimeofday( &t1 , NULL );
   if ( gettimeofdayretcode == 0 ) {
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: generate_announce(): gettimeofday: %d secs, %d usecs.\n", (int)(t1.tv_sec) , (int)(t1.tv_usec) );
      }
   } else {
      fprintf( stderr, "ERROR: gettimeofday failed (ret. code %d).\n", gettimeofdayretcode );
      exit(GET_TIME_OF_THE_DAY_ERROR);
   }

   // prepare the local timer
   struct timeval aux;
   aux.tv_sec = TIMEOUT_REQUEST_SECONDS;
   aux.tv_usec = TIMEOUT_REQUEST_USECONDS;
   timeval_add( &t1, &t1, &aux );

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): entering timerslistempty...\n");
   }

   if ( timerslistempty( timersl ) == 1 ) {
      // list empty. set the global timer
      assert(timerrequest == NULL);
      if (DEBUG) {
         printf("DEBUG: w2lan_core.c: generate_announce(): updating timerrequest...\n");
      }
      timerrequest = timerrequestglobal;
      timerrequest -> tv_sec = TIMEOUT_REQUEST_SECONDS;
      timerrequest -> tv_usec = TIMEOUT_REQUEST_USECONDS;
   }

   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): entering timerslistintoaslast...\n");
   }

   timerslistintoaslast( conversationID , &t1 , (timerslist *)&timersl );
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): Timers, done ! \n");
   }


// TODO send announce
   unsigned char *w2lanannouncepacket;

   w2lanannouncepacket = (unsigned char *)malloc(ETH_P_W2LAN_ANNOUNCE_LEN*sizeof(unsigned char));
   assert(w2lanannouncepacket != NULL);

   // Initialize announce packet 
   memcpy( w2lanannouncepacket, "", 1);

   memcpy( w2lanannouncepacket , (u_int8_t *)ether_aton("FF:FF:FF:FF:FF:FF") , ETHER_ADDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket dst:::::::::: [ ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanannouncepacket[i]);
      }
      printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN-1]);
   }

   memcpy( w2lanannouncepacket + ETHER_ADDR_LEN , myMAC , ETHER_ADDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket src:::::::::::: [ ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanannouncepacket[i+ETHER_ADDR_LEN]);
      }
      printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN-1+ETHER_ADDR_LEN]);
   }

   u_int16_t announcetype = htons( ETH_P_W2LAN_ANNOUNCE );
   memcpy( w2lanannouncepacket+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (u_int8_t *)(&announcetype) , 2 );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket type:::::::::::::: [ ");
      printf("0x%2.2X%2.2X ]\n", w2lanannouncepacket[12] , w2lanannouncepacket[13]);
   }

   memcpy( w2lanannouncepacket + ETHER_HDR_LEN , (u_int8_t *)(packet) , ETHER_ADDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket dst:::::::::::: [ ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanannouncepacket[i+ETHER_HDR_LEN]);
      }
      printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN-1+ETHER_HDR_LEN]);
   }

   memcpy( w2lanannouncepacket + ETHER_HDR_LEN + ETHER_ADDR_LEN, (u_int8_t *)(packet+ETHER_ADDR_LEN) , ETHER_ADDR_LEN );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket src:::::::::::: [ ");
      int i;
      for (i=0; i<ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanannouncepacket[i+ETHER_HDR_LEN+ETHER_ADDR_LEN]);
      }
      printf("%2.2x ]\n", w2lanannouncepacket[ETHER_ADDR_LEN+ETHER_ADDR_LEN-1+ETHER_HDR_LEN]);
   }

   memcpy( w2lanannouncepacket+ETHER_HDR_LEN+ETHER_ADDR_LEN+ETHER_ADDR_LEN , (u_int8_t *)(packet+ETHER_ADDR_LEN+ETHER_ADDR_LEN) , 2 );
   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket type:::::::::::::: [ ");
      printf("0x%2.2X%2.2X ]\n", w2lanannouncepacket[ETHER_HDR_LEN+12] , w2lanannouncepacket[ETHER_HDR_LEN+13]);
   }

   memcpy( w2lanannouncepacket+ETHER_HDR_LEN+ETHER_ADDR_LEN+ETHER_ADDR_LEN+2 , conversationID , ETH_P_W2LAN_CONVERSATIONID_LEN );
   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): w2lanannouncepacket conversationID (%d bytes): [ ", ETH_P_W2LAN_CONVERSATIONID_LEN );
      int i;
      for (i=0; i< ETHER_ADDR_LEN-1; ++i) {
        printf("%2.2x:", w2lanannouncepacket[i+ETHER_HDR_LEN+ETHER_HDR_LEN]);
      }
      printf("%2.2x ]\n", w2lanannouncepacket[ETHER_HDR_LEN+ETHER_HDR_LEN+ETHER_ADDR_LEN-1]);
   }


   if (DEBUG) {
      printf("DEBUG: w2lan_core.c: generate_announce(): pd -> %d\n", (int)pd );
   }

   if (DEBUG) {
      printf( "DEBUG: w2lan_core.c: generate_announce(): packet to sendpacket (%d bytes): [ ", ETH_P_W2LAN_ANNOUNCE_LEN);
      int i;
      for (i=0; i<ETH_P_W2LAN_ANNOUNCE_LEN-1; ++i) {
         printf("%2.2x:", w2lanannouncepacket[i]);
      }
      printf( "%2.2x ]\n", w2lanannouncepacket[ETH_P_W2LAN_ANNOUNCE_LEN-1] );
   }

   if ( ( pcap_sendpacket_retcode = pcap_sendpacket( pd , w2lanannouncepacket , ETH_P_W2LAN_ANNOUNCE_LEN ) ) < 0 ) { 
      fprintf(stderr,"\nError sending the packet: %s\n", pcap_geterr( pd ));
   }

   if (DEBUG) {
     printf("DEBUG: w2lan_core.c: generate_announce(): pcap_sendpacket pcap_sendpacket_retcode: %d (should be >= 0)\n", pcap_sendpacket_retcode );
   }

   free( w2lanannouncepacket );
   free( ethernetframe );
   return;

}

