/***************************************************************************
 *   Copyright (C) 2007 by Francesc Burrull i Mestres.                     *
 *   francesc.burrull@upct.es                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _W2LAN_LISTS_H_
#define _W2LAN_LISTS_H_

#ifndef _W2LAN_H_
#include "w2lan.h"
#endif



/* conversationslist. It is a buffer of SIZEOF_CONVERSATIONSLIST elements */
/* each element is a conversation, consisting of:                         */
/*   conversationID: a six bytes conversation identifier                  */
/*   pending: a boolean indicating if the conversation is in process      */
/*   an ethernet header (mac dst, mac src, data type)                     */


/* number of elements of the list ( default 100 )*/
#define SIZE_CONVERSATIONSLIST 10

typedef struct conversationslistElement *conversationslist;

struct conversationslistElement {
    u_int8_t conversationID[ETHER_ADDR_LEN]; /*it should be ETH_P_W2LAN_CONVERSATIONID_LEN :-( */
    int pending; /* 0 not pending. 1 pending */
    struct {
       u_int8_t  MACdst[ETHER_ADDR_LEN];
       u_int8_t  MACsrc[ETHER_ADDR_LEN];
       u_int16_t MACtype;
    } header;
    conversationslist next;
};

conversationslist conversationslistcreate(void);
conversationslist conversationslistsearch( u_int8_t *conversationID , conversationslist list );
void conversationslistadd( u_int8_t *conversationID , u_int8_t * MACdst, u_int8_t * MACsrc ,  u_int16_t MACtype, conversationslist *list );


int conversationpending( conversationslist conv );
void conversationsetpending( conversationslist conv );
void conversationclearpending( conversationslist conv );
u_int8_t* conversationgetmacdst( conversationslist conv );
u_int8_t* conversationgetmacsrc( conversationslist conv );
u_int16_t conversationgetmactype( conversationslist conv );

/**************************************************************************/

typedef struct counterslistElement *counterslist;

struct counterslistElement {
    u_int8_t conversationID[ETHER_ADDR_LEN]; /*it should be ETH_P_W2LAN_CONVERSATIONID_LEN :-( */
    int requests; 
    counterslist next;
};

counterslist counterslistcreate(void);
void counterslistintoaslast( u_int8_t *conversationID , counterslist *list );
int counterslistfirstinqueue( u_int8_t *conversationID , counterslist *list );
counterslist counterslistsearch( u_int8_t *conversationID , counterslist list );
void counterslistincrease( u_int8_t *conversationID , counterslist list );

/**************************************************************************/

typedef struct pendingframeslistElement *pendingframeslist;

struct pendingframeslistElement {
    u_int8_t conversationID[ETHER_ADDR_LEN]; /*it should be ETH_P_W2LAN_CONVERSATIONID_LEN :-( */
    u_int8_t ethernetdatafield[ETHER_MAX_LEN]; /*it should be ETH_P_W2LAN_DATA_MAX_LEN :-( */
    int size;
    pendingframeslist next;
};

pendingframeslist pendingframeslistcreate(void);
void pendingframeslistput( u_int8_t *conversationID , u_int8_t *ethernetdatafield , u_int16_t size , pendingframeslist *list );
pendingframeslist pendingframeslistsearch( u_int8_t *conversationID , pendingframeslist list );
void pendingframeslistextract( u_int8_t *conversationID , u_int8_t *ethernetdatafield , u_int16_t *size, pendingframeslist *list );

/**************************************************************************/

typedef struct timerslistElement *timerslist;

struct timerslistElement {
    u_int8_t conversationID[ETHER_ADDR_LEN]; /*it should be ETH_P_W2LAN_CONVERSATIONID_LEN :-( */
    struct timeval timer;
    int size;
    timerslist next;
};

timerslist timerslistcreate(void);
int timerslistempty(timerslist list);
timerslist timerslistsearch( u_int8_t *conversationID , timerslist list );
void timerslistintoaslast( u_int8_t *conversationID , struct timeval* t , timerslist *list );
void timerslistfirstinqueue( u_int8_t *conversationID , struct timeval* t , timerslist *list );
void timerslistput( u_int8_t *conversationID , u_int8_t *ethernetdatafield , u_int16_t size ,  timerslist *list );
void timerslistsuccessor( u_int8_t *conversationID , struct timeval *t, timerslist list );
void timerslastinqueue( struct timeval *t, timerslist list );

int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y);
void timeval_add (struct timeval *result, struct timeval *x, struct timeval *y );

#endif  /* _W2LAN_LISTS_H_ */
